'''
Author : hupeng
Time : 2021/8/9 11:09 
Description: 
'''
from . import configparse
from kazoo.client import KazooClient


localconfig = configparse.ConfigParser()


class ZKHelper(object):
    def __init__(self, zk_host: str, zk_path: str):
        self._zk = KazooClient(zk_host)
        self.zk_path = zk_path

    def __enter__(self):
        self._zk.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._zk.stop()

    def get(self, path, space=False):
        path = '{}{}'.format(self.zk_path, path)
        value = self._zk.get(path)[0].decode()
        if space:
            return value
        return value.strip()

    def set_properties(self):
        children = self._zk.get_children(self.zk_path)
        for key in children:
            name = key.replace('.', '_')
            setattr(self, name, self.get(key))



'''
Author : hupeng
Time : 2021/8/10 10:19 
Description: 
'''
import time
import multiprocessing
from collections import defaultdict, OrderedDict
from multiprocessing import cpu_count
from concurrent.futures import ThreadPoolExecutor


class MultiPool(object):
    def __init__(self, max_workers=None, initializer=None, initargs=(),
                 maxtasksperchild=None):
        if max_workers is None:
            max_workers = cpu_count()

        multiprocessing.freeze_support()
        self.pool = multiprocessing.Pool(processes=max_workers, initializer=initializer, initargs=initargs,
                                         maxtasksperchild=maxtasksperchild)

    def submit(self, func, args=(), callback=None,
               error_callback=None, **kwargs):
        return self.pool.apply_async(func, args=args, kwds=kwargs, callback=callback,
                                     error_callback=error_callback)

    def close(self):
        return self.pool.close()

    def join(self):
        return self.pool.join()


class ThreadPool(object):
    def __init__(self):
        self.pool = ThreadPoolExecutor()

    def __call__(self, *args, **kwargs):
        return self.pool


thread_pool = ThreadPool()


class GeneralAsync(object):
    __POOL = {'thread': thread_pool, 'process': MultiPool}

    def __init__(self, num_processor=None, mode='thread'):
        self._funcs = []
        self.num_processor = num_processor
        self.mode = mode
        assert mode in self.__POOL, \
            f'mode param must be `thread` or `process`, but given {mode}'

    def __getattr__(self, item):
        pass

    def _pool(self, **kwargs):
        return self.__POOL[self.mode](**kwargs)

    def add_func(self, func, f_name=None, **params):
        '''
        追加并行执行的任务
        :param func: func
        :param params: param1, param2
        :return:
        '''
        assert callable(func), 'func object must be callable'
        params['func'] = func
        params['f_name'] = f_name or func.__name__
        self._funcs.append(params)

    def run(self):
        task = OrderedDict()
        response = defaultdict(list)
        pool = self._pool(max_workers=len(self._funcs))

        for func_info in self._funcs:
            func = func_info.pop('func')
            f_name = func_info.pop('f_name')
            task[f_name] = pool.submit(func, **func_info)

        if self.mode == 'process':
            pool.close()
            pool.join()

        for name, t in task.items():
            result = t.result() if hasattr(t, 'result') else t.get()
            setattr(self, name, result)
            response[name].append(result)
        return response



'''
Author : hupeng
Time : 2021/8/6 14:49 
Description: 
'''
from .check import Params
from .log import MyLogging
from .asyncs import GeneralAsync
from .wrappers import cal_time, log_time
from .response import MyResponse, responser



'''
Author : hupeng
Time : 2021/8/10 10:12 
Description: 
'''
import re
import typing as t


class Filter_Unkeywords(object):
    '''
    一次完成多个字符串替换
    利用正则表达式re的sub方法
    filter = Filter_Unkeywords(
        [r'\triangle',
        r'\square',
        r'\textcircled']
    )
    filter.multiple_replace(string)
    '''

    def __init__(self, r: t.List[str] = None):
        self.adict = {}
        r = r or []
        for i in r:
            self.adict[i] = ''

    def multiple_replace(self, text):
        rx = re.compile('|'.join(map(re.escape, self.adict)))

        def one_xlat(match):
            return self.adict[match.group(0)]

        # 每遇到一次匹配就会调用回调函数
        return rx.sub(one_xlat, text)



'''
Author : hupeng
Time : 2021/8/26 15:00 
Description: 
'''
from flask import Flask, Blueprint
from fairy import BaseService
from fairy.utils import responser

app = Flask(__name__)

from flask_restful import Api
from flask_restful import Resource

bp = Blueprint('bp', __name__)
api = Api(bp)


class server(BaseService):
    def get_response(self, response):
        response.code = 0
        response.message = 'success'
        response.data = {'res': [1, 2, 3]}
        return response


class test(Resource):
    def get(self):
        response = server().make_response()
        return responser(response=response)


api.add_resource(test, '/test')

app.register_blueprint(bp)

if __name__ == '__main__':
    BaseService.init(cfg={}, app=app)
    app.run()



'''
Author : hupeng
Time : 2021/8/6 15:09 
Description: 
''' 
import time
from functools import wraps


def cal_time(msg: str = None):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            t1 = time.time()
            result = func(*args, **kwargs)
            t2 = time.time()
            if msg:
                print("\033[1;32m %s: %s secs. \033[0m" % (msg, t2 - t1))
            else:
                print("\033[1;32m %s running time: %s secs. \033[0m" % (func.__name__, t2 - t1))
            return result

        return inner

    return wrapper


def log_time(time_record, msg: str = None):
    '''
    记录模块时间
    只用于同步模块
    1、当作用于函数体时：
        @log_time('func执行时间')
        def func():
            pass

    2、当想记录函数体中部分逻辑时：
        from okay_ocr_svr.utils.log import time_record

        t = time.time()
        your code processing
        cost = time.time() - t
        time_record['执行时间'] = float('%.3f' % cost)

    并不需要额外写入到日志文件里
    '''
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            t1 = time.time()
            result = func(*args, **kwargs)
            t2 = time.time()
            cost = float('%.3f' % (t2 - t1))
            time_record[msg or func.__name__] = cost
            return result

        return inner

    return wrapper



'''
Author : hupeng
Time : 2021/8/6 14:57 
Description: 
'''
from flask import request, g
from pydantic import BaseModel


class Params(BaseModel):
    requestid: str = None

    @classmethod
    def params(cls):
        args = request.json
        if args is None:
            try:
                args = request.form.to_dict()
            except:
                args = {}
        return args

    @classmethod
    def build(cls):
        params = cls(**cls.params())
        params.requestid = g.requestid
        return params



'''
Author : hupeng
Time : 2021/8/9 11:10 
Description: 
'''
import os
import re
from typing import (
    Any,
    List,
    Dict,
    AnyStr,
)

from cached_property import cached_property

__all__ = ['ConfigParser']


class Str(str):
    BOOLEAN_STATES = {'1': True, 'yes': True, 'true': True, 'on': True,
                      '0': False, 'no': False, 'false': False, 'off': False}
    STRIP = ''' '"'''
    __int = int
    __dict = dict

    @cached_property
    def _bool(self) -> bool:
        _str = self.__str__()
        if _str.lower() not in self.BOOLEAN_STATES:
            raise ValueError('Not a boolean: %s' % _str)
        return self.BOOLEAN_STATES[_str.lower()]

    @cached_property
    def _int(self) -> int:
        return self.__int(self.__str__())

    @cached_property
    def _list(self) -> List[AnyStr]:
        _str = self.__str__()
        if not (_str.startswith('[') and _str.endswith(']')):
            raise ValueError("can't convert list object")
        return [_.strip(self.STRIP) for _ in _str.strip('[]').split(',')]

    @cached_property
    def _dict(self) -> Dict[str, str]:
        _str = self.__str__()
        if not (_str.startswith('{') and _str.endswith('}')):
            raise ValueError("can't convert dict object")
        _l = [_.strip(self.STRIP).split(':', 1) for _ in _str.strip('{}').split(',')]
        return self.__dict(_l)


class Section(object):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        s = f'{self.name} section info:\n'
        for key, value in self.__dict__.items():
            s += f'{key}: {value}\n'
        return s


class BaseConfigParser(object):
    _SECTION_PATTERN = '\[(?P<header>[^]]+)\]'
    _OPT_TMPL = '(?P<option>.*?)\s*(?P<vi>{delim})\s*(?P<value>.*)$'

    SECTCRE = re.compile(_SECTION_PATTERN, re.VERBOSE)
    OPTCRE = re.compile(_OPT_TMPL.format(delim="=|:"), re.VERBOSE)
    _KEYCRE = re.compile(r"%\((?P<key>[^)]+)\)s")

    def before_set(self, section: Section, value: Any) -> Any:
        tmp_value = value.replace('%%', '')  # escaped percent signs
        tmp_value = self._KEYCRE.sub('', tmp_value)  # valid syntax
        if '%' in tmp_value:
            raise ValueError("invalid interpolation syntax in %r at "
                             "position %d" % (value, tmp_value.find('%')))
        replace_k = self._KEYCRE.search(value)
        if replace_k is not None:
            replace_option = replace_k.group('key')
            if not hasattr(section, replace_option):
                raise ValueError(f'section {section.name} not have option {replace_option}')
            value = getattr(section, replace_option) + tmp_value
        return value


class ConfigParser(BaseConfigParser):
    def __init__(self):
        self._sections = {}

    def __getattr__(self, item):
        # print(item)
        return self.section(item)

    @property
    def sections(self):
        return list(self._sections.keys())

    def section(self, name):
        if name in self._sections:
            return self._sections[name]
        raise AttributeError(f"No section: '{name}'")

    def read(self, filenames: str, encoding: str = 'utf-8-sig') -> List[str]:
        if isinstance(filenames, (str, os.PathLike)):
            filenames = [filenames]

        read_ok = []
        for filename in filenames:
            try:
                with open(filename, encoding=encoding) as fp:
                    self._read(fp, filename)
            except OSError:
                continue
            if isinstance(filename, os.PathLike):
                filename = os.fspath(filename)
            read_ok.append(filename)
        return read_ok

    def _read(self, fp: open, fpname: str) -> None:
        lines = [line for line in fp.readlines() if line.strip() and (not line.strip().startswith(';'))
                 and (not line.strip().startswith('#'))]
        if not lines:
            raise (f'{fpname} can not be empty!')
        section = Section(self.SECTCRE.match(lines[0]).group('header'))
        sections = [section]
        self.sections.append(section.name)
        # print(section.name)
        for line in lines[1:]:
            sec_str = self.SECTCRE.match(line)
            if sec_str:
                sec_name = sec_str.group('header')
                section = Section(sec_name)
                sections.append(section)
                # self.sections.append(section.name)

            opt = self.OPTCRE.search(line)
            if opt:
                option, _, value = opt.groups()
                value = self.before_set(section, value)
                # print(option, value)
                setattr(section, option, Str(value))

        for sec in sections:
            self._sections[sec.name] = sec
            # setattr(self, sec.name, sec)


if __name__ == '__main__':
    config = ConfigParser()
    config.read('config.ini')
    # print(config.sections)
    print(config.log.log_dir)
    print(config.db.port._int, type(config.db.port._int))
    print(config.server.debug._bool, type(config.server.debug._bool))
    print(config.oxk.high._list, type(config.oxk.high._list))
    print(config.oxk.high._list, type(config.oxk.high._list))
    print(config.oxk.high._list, type(config.oxk.high._list))
    # print(config.l.di._dict())



'''
Author : hupeng
Time : 2021/8/6 14:47 
Description: 
'''
import json
import traceback

from flask import request, Flask, g, Response

from fairy.const import Code
from fairy.utils import MyResponse
from fairy.utils.log import rd, ErrorLog
from fairy.utils.log import logger, MyLogging


class BaseConfig(object):
    def __init__(self):
        self.data = {}

    def dumps(self):
        return json.dumps(self.data)

    def __setitem__(self, key, value):
        self.data[key] = value

    def __getitem__(self, item):
        return self.data[item]

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return str(self.data)


error_log = server_log = logger


def before_request():
    url = request.url
    args = request.form or request.json
    g.args = args
    g.requestid = request.headers.get('requestid')
    rd.requestid = g.requestid
    if not g.requestid:
        data = {'code': 10000, 'message': 'miss requestid', 'data': {}}
        return Response(
            response=json.dumps(data, ensure_ascii=False),
            mimetype='application/json')
    try:
        server_log.write('[request] [path: %s] %s' % (request.path, str(args)))
    except Exception:
        error_log.write("request url: %s, response: message too long" % url)


def after_request(response):
    url = request.url
    resp = response.data.decode()
    try:
        resps = json.loads(resp)
        if resps['code'] != 0:
            return response
    except Exception:
        pass
    try:
        server_log.write('[response] %s' % resp)
    except Exception:
        error_log.write("request url: %s, response: message too long" % url)
    return response


class BaseService(object):
    config = None
    params_class = None
    __is_init = False

    @classmethod
    def init(cls, cfg: dict, name: str = 'server', app: Flask = None, **options):
        print('init...')
        cls.config = cfg
        cls.CODE_SUCCESS = cfg.get('CODE_SUCCESS', Code(0, '成功'))
        cls.CODE_PARAM = cfg.get('CODE_PARAM', Code(10000, '参数错误'))
        cls.CODE_SERVICE = cfg.get('CODE_SERVICE', Code(10001, '服务异常'))
        cls.__is_init = True
        use_log = options.get('use_log', True)
        if use_log:
            cls.global_logger(name)
        use_hook = options.get('use_hook', True)
        if use_hook and app:
            cls.init_hooks(app)

    @classmethod
    def global_logger(cls, name):
        global error_log
        global server_log
        loggers = cls.config.get('loggers', {})
        log_dir = cls.config.get('log_dir', './log')
        error = loggers.get('error')
        if error is None:
            error = ErrorLog(log_dir)
        error_log = error
        server = loggers.get('server')
        if server is None:
            server = MyLogging(log_dir, name, name)
        server_log = server

    @classmethod
    def init_hooks(cls, app):
        hooks = cls.config.get('hooks', {})
        before_request_func = hooks.get(
            'before_request_func',
            before_request
        )
        after_request_func = hooks.get(
            'after_request_func',
            after_request
        )
        app.before_request(before_request_func)
        app.after_request(after_request_func)

    def __init__(self):
        self._request = request
        self.params = None

    def get_params(self):
        if self.params_class is None:
            return
        params = self.params_class.build()
        self.params = params
        return params

    def prints(self, text: str, p=True, c=False):
        if p and c:
            print('\033[0;32m{}\033[0m'.format(text))
        elif p:
            print(text)

    def make_response(self) -> MyResponse:
        response = MyResponse()
        # 参数校验
        try:
            if not self.__is_init:
                self.init({})
            self.get_params()
        except Exception as e:
            response.code = self.CODE_PARAM.val
            response.message = str(e)
            error_log.write('got a params error %s' %
                            str(traceback.format_exc()))
            return response

        response.code = self.CODE_SUCCESS.val
        response.message = self.CODE_SUCCESS.msg
        try:
            response = self.get_response(response)
        except Exception as e:
            response.code = self.CODE_SERVICE.val
            response.message = f'{self.CODE_SERVICE.msg}: {str(e)}'
            error_log.write(str(traceback.format_exc()))
        return response

    def get_response(self, response: MyResponse) -> MyResponse:
        raise NotImplementedError()



'''
Author : hupeng
Time : 2021/8/6 14:27 
Description: 
'''
import setuptools
from fairy import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fairy_tail",
    version=__version__,
    author="BenjyHu",
    license="MIT",
    author_email="m15227041551@163.com",
    description="A small smart package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BenjyHu/fairy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'flask',
        'kazoo',
        'pydantic',
        'cached_property'
    ],
    python_requires=">=3.6",
)
